@@ -17,6 +17,7 @@ from nubrain.experiment.randomize_conditions import (
     sample_next_image,
     shuffle_with_repetitions,
 )
+from nubrain.experiment_eeg_to_image_v1.tone import generate_tone
 from nubrain.image.tools import get_all_images, load_and_scale_image
 from nubrain.misc.datetime import get_formatted_current_datetime
 
@@ -87,12 +88,10 @@ def experiment_eeg_to_image_v1(config: dict):
     image_categories = list(set([x["image_category"] for x in images_and_categories]))
     random.shuffle(image_categories)
 
-    n_trials = n_blocks * images_per_block
-
     # Order of image categories.
     trial_order = create_balanced_list(
         image_categories=image_categories,
-        target_length=n_trials,
+        target_length=n_blocks,
     )
     random.shuffle(trial_order)
 
@@ -191,7 +190,7 @@ def experiment_eeg_to_image_v1(config: dict):
         # Timing parameters
         "initial_rest_duration": initial_rest_duration,
         "image_duration": image_duration,
-        "isi_duration": isi_duration,
+        "isi_duration": 0.0,  # Dummy value. TODO: Adjust logging for live eeg-to-image
         "inter_block_grey_duration": inter_block_grey_duration,
         # Experiment structure
         "n_blocks": n_blocks,
@@ -213,6 +212,35 @@ def experiment_eeg_to_image_v1(config: dict):
     while running:
         pygame.init()
 
+        # ------------------------------------------------------------------------------
+        # *** Prepare audio cue
+
+        pygame.mixer.init(frequency=44100, size=-16, channels=2, buffer=512)
+
+        frequency = 1000  # Pitch of the tone in Hz (e.g., 1000 Hz is a common beep)
+        duration = 0.3  # Duration in seconds (half a second)
+        amplitude = 0.5  # Volume, from 0.0 to 1.0
+
+        # Play the tone x seconds before the end of the pre-stimulus period.
+        tone_pre_stimulus_onset = 0.6
+
+        # Get the sample rate from the mixer settings.
+        sample_rate = pygame.mixer.get_init()[0]
+
+        # Generate the tone data.
+        tone_data = generate_tone(
+            frequency=frequency,
+            duration=duration,
+            amplitude=amplitude,
+            sample_rate=sample_rate,
+        )
+
+        # Create a Sound object from the numpy array.
+        pure_tone = pygame.sndarray.make_sound(tone_data)
+
+        # ------------------------------------------------------------------------------
+        # *** Prepare visual stimulus generation
+
         # Get screen dimensions and set up full screen.
         screen_info = pygame.display.Info()
         screen_width = screen_info.current_w
@@ -223,8 +251,6 @@ def experiment_eeg_to_image_v1(config: dict):
         pygame.display.set_caption("Image Presentation Experiment")
         pygame.mouse.set_visible(False)
 
-        idx_trial = 0
-
         try:
             # Initial grey screen.
             pygame.time.wait(100)
@@ -235,13 +261,50 @@ def experiment_eeg_to_image_v1(config: dict):
             pygame.display.flip()
 
             # Pause for specified number of milliseconds.
-            pygame.time.delay(int(round(initial_rest_duration * 1000.0)))
+            pygame.time.delay(
+                int(round((initial_rest_duration - tone_pre_stimulus_onset) * 1000.0))
+            )
 
             # Clear board buffer.
             _, _ = eeg_device.get_board_data()
 
             # Block loop.
             for idx_block in range(n_blocks):
+                # Average embedding vectors within blocks (across trials). Show x
+                # repetitions of the same image, perform inference, and average the
+                # embedding vector.
+
+                # Sample the next image.
+                next_image_category = trial_order[idx_block]
+                next_image_file_path = sample_next_image(
+                    next_image_category=next_image_category,
+                    category_to_filepath=category_to_filepath,
+                    previous_image_file_path=previous_image_file_path,
+                )
+
+                # Load the next image.
+                image_and_metadata = None
+                while image_and_metadata is None:
+                    image_and_metadata = load_and_scale_image(
+                        image_file_path=next_image_file_path,
+                        screen_width=screen_width,
+                        screen_height=screen_height,
+                    )
+
+                current_image = image_and_metadata["image"]
+
+                # Play tone to cue block start.
+                pure_tone.play()
+                pygame.time.delay(int(round(tone_pre_stimulus_onset * 1000.0)))
+
+                eeg_data = []
+                eeg_timestamps = []
+
+                # `marker_data` is an array of shape [2, n_timepoints], where the [0, :]
+                # corresponds to timestamps, and [1, :] corresponds to the marker values
+                # (represented as a nested list here).
+                marker_data = np.zeros((2, (images_per_block * 2)))
+
                 # Image loop (within a block).
                 for image_count in range(images_per_block):
                     if not running:  # Check for quit event
@@ -253,32 +316,13 @@ def experiment_eeg_to_image_v1(config: dict):
                     # Start of the pre-stimulus interval.
                     t_pre_stim_start = time()
 
-                    # Sample the next image.
-                    next_image_category = trial_order[idx_trial]
-                    next_image_file_path = sample_next_image(
-                        next_image_category=next_image_category,
-                        category_to_filepath=category_to_filepath,
-                        previous_image_file_path=previous_image_file_path,
-                    )
-
-                    # Load the next image.
-                    image_and_metadata = None
-                    while image_and_metadata is None:
-                        image_and_metadata = load_and_scale_image(
-                            image_file_path=next_image_file_path,
-                            screen_width=screen_width,
-                            screen_height=screen_height,
-                        )
-
-                    current_image = image_and_metadata["image"]
-
                     img_rect = current_image.get_rect(
                         center=(screen_width // 2, screen_height // 2)
                     )
                     screen.fill(global_config.rest_condition_color)
                     screen.blit(current_image, img_rect)
 
-                    # Wait until the end of the pre-stimulus period
+                    # Wait until the end of the pre-stimulus period.
                     t_pre_stim_end = t_pre_stim_start + pre_stimulus_interval
                     while time() < t_pre_stim_end:
                         for event in pygame.event.get():
@@ -310,6 +354,8 @@ def experiment_eeg_to_image_v1(config: dict):
                                 "timestamp": marker_ts_stim_start,
                             }
                         )
+                    marker_data[0, (image_count * 2)] = marker_ts_stim_start
+                    marker_data[1, (image_count * 2)] = marker_val_stim_start
 
                     # Send pre-stimulus EEG data (to avoid buffer overflow).
                     eeg_data_pre_stim, eeg_ts_pre_stim = eeg_device.get_board_data()
@@ -321,6 +367,8 @@ def experiment_eeg_to_image_v1(config: dict):
                                 "eeg_timestamps": eeg_ts_pre_stim,
                             }
                         )
+                    eeg_data.append(eeg_data_pre_stim)
+                    eeg_timestamps.append(eeg_ts_pre_stim)
 
                     # Wait for image duration, but check for responses continuously.
                     t_stim_end_expected = t_stim_start + image_duration
@@ -355,6 +403,8 @@ def experiment_eeg_to_image_v1(config: dict):
                                 "timestamp": marker_ts_stim_end,
                             }
                         )
+                    marker_data[0, ((image_count * 2) + 1)] = marker_ts_stim_end
+                    marker_data[1, ((image_count * 2) + 1)] = marker_val_stim_end
 
                     eeg_data_stim, eeg_ts_stim = eeg_device.get_board_data()
                     if eeg_data_stim.size > 0:
@@ -365,6 +415,8 @@ def experiment_eeg_to_image_v1(config: dict):
                                 "eeg_timestamps": eeg_ts_stim,
                             }
                         )
+                    eeg_data.append(eeg_data_stim)
+                    eeg_timestamps.append(eeg_ts_stim)
 
                     # Wait until the end of the post-stimulus period
                     t_post_stim_end = t_stim_end_actual + post_stimulus_interval
@@ -389,6 +441,8 @@ def experiment_eeg_to_image_v1(config: dict):
                                 "eeg_timestamps": eeg_ts_post_stim,
                             }
                         )
+                    eeg_data.append(eeg_data_post_stim)
+                    eeg_timestamps.append(eeg_ts_post_stim)
 
                     # ------------------------------------------------------------------
                     # *** Log stimulus metadata
@@ -399,177 +453,169 @@ def experiment_eeg_to_image_v1(config: dict):
                         "stimulus_duration_s": t_stim_end_actual - t_stim_start,
                         "image_file_path": next_image_file_path,
                         "image_category": next_image_category,
+                        "is_target_event": False,  # Dummy values, TODO: data logging for eeg-to-image
+                        "response_time_s": 0.0,  # Dummy values, TODO: data logging for eeg-to-image
                     }
                     data_logging_queue.put(
                         {"type": "stimulus", "stimulus_data": stimulus_data}
                     )
 
-                    # ------------------------------------------------------------------
-                    # *** Inference
+                # ----------------------------------------------------------------------
+                # *** Inference
 
-                    eeg_data = np.concatenate(
-                        [eeg_data_pre_stim, eeg_data_stim, eeg_data_post_stim],
-                        axis=1,
+                eeg_data = np.concatenate(eeg_data, axis=1)
+
+                eeg_timestamps = np.concatenate(eeg_timestamps)
+
+                request_dict = {
+                    "eeg_data": eeg_data.tolist(),
+                    "eeg_timestamps": eeg_timestamps.tolist(),
+                    "marker_data": marker_data.tolist(),
+                    "utility_frequency": utility_frequency,
+                    "eeg_channel_mapping": eeg_channel_mapping,
+                }
+
+                request_json = json.dumps(request_dict)
+
+                response = requests.post(
+                    url=api_endpoint,
+                    data=request_json,
+                )
+
+                time_now = get_formatted_current_datetime()
+                true_image_category = next_image_category
+
+                if response.status_code == 200:
+                    eeg_model_id = response.headers.get("X-eeg-model-id")
+
+                    path_image_out = os.path.join(
+                        output_dir_images,
+                        f"{eeg_model_id}_{time_now}_{true_image_category}.png",
                     )
 
-                    eeg_timestamps = np.concatenate(
-                        [eeg_ts_pre_stim, eeg_ts_stim, eeg_ts_post_stim]
+                    with open(path_image_out, "wb") as f:
+                        f.write(response.content)
+                else:
+                    print(f"Error: {response.status_code}")
+                    print(response.text)
+                    running = False
+                    break
+
+                # ----------------------------------------------------------------------
+                # *** Show generated image
+
+                generated_image_and_metadata = load_and_scale_image(
+                    image_file_path=path_image_out,
+                    screen_width=screen_width,
+                    screen_height=screen_height,
+                )
+
+                generated_image = generated_image_and_metadata["image"]
+
+                img_rect = current_image.get_rect(
+                    center=(screen_width // 2, screen_height // 2)
+                )
+                screen.fill(global_config.rest_condition_color)
+                screen.blit(generated_image, img_rect)
+                pygame.display.flip()
+
+                # Show generated image for this amount of time.
+                t_generated_img_end = time() + generated_image_duration
+
+                # Insert stimulus start marker and get its timestamp.
+                generated_img_start_marker = 3.0  # Hardcoded TODO make config param
+                marker_val_stim_start, marker_ts_stim_start = eeg_device.insert_marker(
+                    generated_img_start_marker
+                )
+                if marker_val_stim_start is not None:
+                    data_logging_queue.put(
+                        {
+                            "type": "marker",
+                            "marker_value": marker_val_stim_start,
+                            "timestamp": marker_ts_stim_start,
+                        }
                     )
 
-                    # `marker_data` is an array of shape [2, n_timepoints], where the
-                    # [0, :] corresponds to timestamps, and [1, :] corresponds to the
-                    # marker values (represented as a nested list here).
-                    marker_data = np.array(
-                        [
-                            [marker_ts_stim_start, marker_ts_stim_end],
-                            [marker_val_stim_start, marker_val_stim_end],
-                        ]
-                    )
-
-                    request_dict = {
-                        "eeg_data": eeg_data.tolist(),
-                        "eeg_timestamps": eeg_timestamps.tolist(),
-                        "marker_data": marker_data.tolist(),
-                        "utility_frequency": utility_frequency,
-                        "eeg_channel_mapping": eeg_channel_mapping,
-                    }
-
-                    request_json = json.dumps(request_dict)
-
-                    response = requests.post(
-                        url=api_endpoint,
-                        data=request_json,
-                    )
-
-                    time_now = get_formatted_current_datetime()
-                    true_image_category = next_image_category
-
-                    if response.status_code == 200:
-                        eeg_model_id = response.headers.get("X-eeg-model-id")
-
-                        path_image_out = os.path.join(
-                            output_dir_images,
-                            f"{eeg_model_id}_{time_now}_{true_image_category}.png",
-                        )
-
-                        with open(path_image_out, "wb") as f:
-                            f.write(response.content)
-                    else:
-                        print(f"Error: {response.status_code}")
-                        print(response.text)
-                        running = False
-                        break
-
-                    # ------------------------------------------------------------------
-                    # *** Log EEG data (to avoid buffer overflow)
-
-                    # Log data from the interval when waiting for inference results.
-
-                    eeg_data_wait, eeg_ts_wait = eeg_device.get_board_data()
-                    if eeg_data_wait.size > 0:
-                        data_logging_queue.put(
-                            {
-                                "type": "eeg",
-                                "eeg_data": eeg_data_wait,
-                                "eeg_timestamps": eeg_ts_wait,
-                            }
-                        )
-
-                    # ------------------------------------------------------------------
-                    # *** Show generated image
-
-                    generated_image_and_metadata = load_and_scale_image(
-                        image_file_path=path_image_out,
-                        screen_width=screen_width,
-                        screen_height=screen_height,
-                    )
-
-                    generated_image = generated_image_and_metadata["image"]
-
-                    img_rect = current_image.get_rect(
-                        center=(screen_width // 2, screen_height // 2)
-                    )
-                    screen.fill(global_config.rest_condition_color)
-                    screen.blit(generated_image, img_rect)
-
-                    # Show generated image for this amount of time. Includes the jitter.
-                    t_generated_img_end = (
-                        time()
-                        + generated_image_duration
-                        + np.random.uniform(low=0.0, high=isi_jitter)
-                    )
-
-                    # Insert stimulus start marker and get its timestamp.
-                    generated_img_start_marker = 3.0  # Hardcoded TODO make config param
-                    marker_val_stim_start, marker_ts_stim_start = (
-                        eeg_device.insert_marker(generated_img_start_marker)
-                    )
-                    if marker_val_stim_start is not None:
-                        data_logging_queue.put(
-                            {
-                                "type": "marker",
-                                "marker_value": marker_val_stim_start,
-                                "timestamp": marker_ts_stim_start,
-                            }
-                        )
-
-                    while time() < t_generated_img_end:
-                        for event in pygame.event.get():
-                            if event.type == pygame.QUIT:
+                while time() < t_generated_img_end:
+                    for event in pygame.event.get():
+                        if event.type == pygame.QUIT:
+                            running = False
+                        if event.type == pygame.KEYDOWN:
+                            if event.key == pygame.K_ESCAPE:
                                 running = False
-                            if event.type == pygame.KEYDOWN:
-                                if event.key == pygame.K_ESCAPE:
-                                    running = False
-                        if not running:
-                            break
                     if not running:
                         break
+                if not running:
+                    break
 
-                    # ------------------------------------------------------------------
-                    # *** Grey screen (after generated image)
+                # ----------------------------------------------------------------------
+                # *** Log EEG data (to avoid buffer overflow)
 
-                    # End of generated image presentation. Display grey screen.
-                    screen.fill(global_config.rest_condition_color)
-                    pygame.display.flip()
+                # Log data from the interval when waiting for inference results.
 
-                    # Insert stimulus start marker and get its timestamp.
-                    generated_img_end_marker = 4.0  # Hardcoded TODO make config param
-                    marker_val_stim_end, marker_ts_stim_end = eeg_device.insert_marker(
-                        generated_img_end_marker
+                eeg_data_wait, eeg_ts_wait = eeg_device.get_board_data()
+                if eeg_data_wait.size > 0:
+                    data_logging_queue.put(
+                        {
+                            "type": "eeg",
+                            "eeg_data": eeg_data_wait,
+                            "eeg_timestamps": eeg_ts_wait,
+                        }
                     )
-                    if marker_val_stim_end is not None:
-                        data_logging_queue.put(
-                            {
-                                "type": "marker",
-                                "marker_value": marker_val_stim_end,
-                                "timestamp": marker_ts_stim_end,
-                            }
-                        )
 
-                    # ------------------------------------------------------------------
-                    # *** Log EEG data (to avoid buffer overflow)
+                # ----------------------------------------------------------------------
+                # *** Grey screen (after generated image)
 
-                    eeg_data_gen_img, eeg_ts_gen_img = eeg_device.get_board_data()
-                    if eeg_data_gen_img.size > 0:
-                        data_logging_queue.put(
-                            {
-                                "type": "eeg",
-                                "eeg_data": eeg_data_gen_img,
-                                "eeg_timestamps": eeg_ts_gen_img,
-                            }
-                        )
+                # End of generated image presentation. Display grey screen.
+                screen.fill(global_config.rest_condition_color)
+                pygame.display.flip()
 
-                    # ------------------------------------------------------------------
-                    # *** Prepare next trial
+                # Insert stimulus start marker and get its timestamp.
+                generated_img_end_marker = 4.0  # Hardcoded TODO make config param
+                marker_val_stim_end, marker_ts_stim_end = eeg_device.insert_marker(
+                    generated_img_end_marker
+                )
+                if marker_val_stim_end is not None:
+                    data_logging_queue.put(
+                        {
+                            "type": "marker",
+                            "marker_value": marker_val_stim_end,
+                            "timestamp": marker_ts_stim_end,
+                        }
+                    )
 
-                    # Update tracking variables for the next loop iteration.
-                    previous_image_file_path = next_image_file_path
+                # ----------------------------------------------------------------------
+                # *** Log EEG data (to avoid buffer overflow)
 
-                    idx_trial += 1
+                eeg_data_gen_img, eeg_ts_gen_img = eeg_device.get_board_data()
+                if eeg_data_gen_img.size > 0:
+                    data_logging_queue.put(
+                        {
+                            "type": "eeg",
+                            "eeg_data": eeg_data_gen_img,
+                            "eeg_timestamps": eeg_ts_gen_img,
+                        }
+                    )
 
                 if not running:
                     break
 
+                # ----------------------------------------------------------------------
+                # *** Prepare next block
+
+                # Update tracking variables for the next loop iteration.
+                previous_image_file_path = next_image_file_path
+
+                # Inter-block grey screen.
+                screen.fill(global_config.rest_condition_color)
+                pygame.display.flip()
+                remaining_wait = (
+                    inter_block_grey_duration
+                    - tone_pre_stimulus_onset
+                    + np.random.uniform(low=0.0, high=isi_jitter)
+                )
+                pygame.time.delay(int(round(remaining_wait * 1000.0)))
+
                 # Send post-stimulus EEG data (to avoid buffer overflow).
                 eeg_data, eeg_ts = eeg_device.get_board_data()
                 if eeg_data.size > 0:
@@ -577,14 +623,8 @@ def experiment_eeg_to_image_v1(config: dict):
                         {"type": "eeg", "eeg_data": eeg_data, "eeg_timestamps": eeg_ts}
                     )
 
-                # Inter-block grey screen.
-                screen.fill(global_config.rest_condition_color)
-                pygame.display.flip()
-                # We already waited for the ISI duration, therefore subtract it from the
-                # inter block duration. Avoid negative value in case ISI duration is
-                # longer than inter block duration.
-                remaining_wait = max((inter_block_grey_duration - isi_duration), 0.0)
-                pygame.time.delay(int(round(remaining_wait * 1000.0)))
+            # --------------------------------------------------------------------------
+            # *** End of experiment
 
             running = False
 
